# coding: windows-1251
"""
PIGPASS section
"""
from .ordered_attrib import ET
from .codes import PassType
from .base import Section as InfotechSection, AbstractItem


class Item(AbstractItem):
    """
    object item from <PIGPASS> xml section
    <PASS
      IDTYPEOBJ="0" DATE1="25.05.2018 10:58:00" DATE2="25.05.2018 11:10:00"
      SPEED_AVERAGE="0.26" REM="" MANUFACTURER="xxxx" MANUFACT_DATE="2010"
      PIGTYPE="990005096296" OBSLTYPE="2"
    />
    """
    xml_node_name = 'PASS'

    field_typeobj = 'IDTYPEOBJ'
    field_date1 = 'DATE1'
    field_date2 = 'DATE2'
    field_speed = 'SPEED_AVERAGE'
    field_rem = 'REM'
    field_manufacturer = 'MANUFACTURER'
    field_manufac_date = 'MANUFACT_DATE'
    field_pigtype = 'PIGTYPE'
    field_insptype = 'OBSLTYPE'

    def __init__(self):
        self.pig = None
        self.objtype = None
        self.date1 = None
        self.date2 = None
        self.speed = None
        self.rem = None
        self.manufacturer = None
        self.manufac_date = None
        self.pigtype = None
        self.insptype = None

    @classmethod
    def from_xml(cls, xml_item):
        """
        create item from xml item
        """
        obj = cls()

        obj.objtype = xml_item.attrib[Item.field_typeobj]
        obj.date1 = xml_item.attrib[Item.field_date1]
        obj.date2 = xml_item.attrib[Item.field_date2]
        obj.speed = xml_item.attrib[Item.field_speed]
        obj.rem = xml_item.attrib[Item.field_rem]
        obj.manufacturer = xml_item.attrib[Item.field_manufacturer]
        obj.manufac_date = xml_item.attrib[Item.field_manufac_date]
        obj.pigtype = xml_item.attrib[Item.field_pigtype]
        obj.insptype = xml_item.attrib[Item.field_insptype]

        return obj

    def add_xml_child(self, parent_node):
        """
        create and add pigpass xml node of object to parent xml node
        """
        node = ET.SubElement(parent_node, Item.xml_node_name)

        node.set(Item.field_typeobj, self.objtype)
        node.set(Item.field_date1, self.date1)
        node.set(Item.field_date2, self.date2)
        node.set(Item.field_speed, self.speed)
        node.set(Item.field_rem, self.rem)
        node.set(Item.field_manufacturer, self.manufacturer)
        node.set(Item.field_manufac_date, self.manufac_date)
        node.set(Item.field_pigtype, self.pigtype)
        node.set(Item.field_insptype, self.insptype)

        return node


class Section(InfotechSection):
    """
    <PIGPASS> xml section
    """
    section_tag = 'PIGPASS'
    item_attributes = [
      Item.field_typeobj,
      Item.field_date1,
      Item.field_date2,
      Item.field_speed,
      Item.field_rem,
      Item.field_manufacturer,
      Item.field_manufac_date,
      Item.field_pigtype,
      Item.field_insptype,
    ]

    def __init__(self, infotech):
        super(Section, self).__init__(infotech, Item, Section.section_tag)

    def is_navigate(self):
        """
        is navigate data present
        """
        if not self._items:
            return False

        for pigpass in self._items:
            if pigpass.insptype in [PassType.NAVIGATE, PassType.COMPLEX_NAV]:
                return True

        return False



"""
WELDS section
"""
from .ordered_attrib import ET
from .base import DistItem, Section as InfotechSection, to_int
from . import reverse_orient


class Item(DistItem):
    """
    weld item from <WELDS> xml section
    <WLD
      IDTYPEOBJ="2097791" ODOMETER="0"
      NUM_TUBE="1" DL_TUBE="752" THICK="19.5" PSH1="11.6" PSH2="" REM=""
    />
    """
    xml_node_name = 'WLD'
    field_number = 'NUM_TUBE'
    field_length = 'DL_TUBE'
    field_thick = 'THICK'
    field_hor1 = 'PSH1'
    field_hor2 = 'PSH2'

    def __init__(self):
        super(Item, self).__init__()

        self.number = ''
        self.length = ''
        self.thick = ''
        self.hor1 = ''
        self.hor2 = ''

    @classmethod
    def from_xml(cls, xml_item):
        """
        create weld item from existing xml element
        """
        obj = cls()
        obj.fill_from_xml(xml_item)

        obj.number = xml_item.attrib[Item.field_number]
        obj.length = to_int(xml_item.attrib[Item.field_length])
        obj.thick = xml_item.attrib[Item.field_thick]
        obj.hor1 = xml_item.attrib[Item.field_hor1]
        obj.hor2 = xml_item.attrib[Item.field_hor2]

        return obj

    def end(self):
        """
        return distance for end of tube
        """
        return self.dist + self.length

    def as_csv_row(self, infotech, with_navigation=False):
        """
        return list of field values for csv string
        """
        columns = [
          self.number,
          "{}".format(self.length),
          self.thick.replace('.', ','),
          self.hor1.replace('.', ','),
          self.hor2.replace('.', ','),
        ]
        base_columns = super(Item, self).as_csv_row(infotech, with_navigation=with_navigation)

        return base_columns[:2] + columns + base_columns[2:]

    def reverse(self, total_length, object_index):
        """
        reverse weld
        """
        self.dist = total_length - (self.dist + self.length)
        self.number = "{}".format(object_index)
        self.hor1 = reverse_orient(self.hor1)
        self.hor2 = reverse_orient(self.hor2)

    def add_xml_child(self, parent_node):
        """
        create and add xmml node of weld to parent xml node
        """
        node = ET.SubElement(parent_node, Item.xml_node_name)
        super(Item, self).base_xml(node)
        node.set(Item.field_number, self.number)
        node.set(Item.field_length, "{}".format(self.length))
        node.set(Item.field_thick, self.thick)
        node.set(Item.field_hor1, self.hor1)
        node.set(Item.field_hor2, self.hor2)

        return node


class Section(InfotechSection):
    """
    <WELDS> xml section
    """
    item_attributes = DistItem.dist_attribs + [
      Item.field_number,
      Item.field_length,
      Item.field_thick,
      Item.field_hor1,
      Item.field_hor2,
      DistItem.field_comment,
    ] + DistItem.coords_attribs

    def __init__(self, infotech):
        super(Section, self).__init__(infotech, Item, 'WELDS')

    def as_csv(self, with_navigation=False):
        """
        dump weld section content as csv string
        """
        column_titles = [
          'Name', 'Distance', 'Tube number', 'Tube length', 'Thick', 'Weld1', 'Weld2', 'Comment'
        ]
        return super(Section, self).as_csv_body('Welds table', column_titles, with_navigation=with_navigation)

    def add_tube(self, objtype, tube_dist, tube_length):
        """
        add new tube with given type, dist and length to thr end of record
        """
        item = Item()

        item.objtype = objtype
        item.dist = tube_dist
        item.comment = ''

        item.number = ''
        item.length = tube_length
        item.thick = self.items[-1].thick
        item.hor1 = ''
        item.hor2 = ''

        self._items.append(item)



"""
DEFECTS section
"""
from .ordered_attrib import ET
from .base import Section as InfotechSection, DistItem, to_int
from . import reverse_orient


class Item(DistItem):  # pylint: disable=too-many-instance-attributes
    """
    defect item from <DEFECTS> xml section
    <DEF
      IDTYPEOBJ="990004698869" ODOMETER="145"
      L_OTCH="170" W_OTCH="309" V_MIN_OTCH="5" V_MAX_OTCH="5" ORIENT1="7.2" ORIENT2="8.1" NUMDEF="2" REM=""
      KBD="0.70" PBEZ="" TIME_LIMIT="23.686" PBEZ_PERCENT="" METHOD="6907370"
    />
    """
    xml_node_name = 'DEF'
    field_length = 'L_OTCH'
    field_width = 'W_OTCH'
    field_loss_min = 'V_MIN_OTCH'
    field_loss_max = 'V_MAX_OTCH'
    field_orient_start = 'ORIENT1'
    field_orient_end = 'ORIENT2'
    field_number = 'NUMDEF'
    field_kbd = 'KBD'
    field_safe_pressure = 'PBEZ'
    field_time_limit = 'TIME_LIMIT'
    field_safe_pressure_persent = 'PBEZ_PERCENT'
    field_method_id = 'METHOD'

    def __init__(self):
        super(Item, self).__init__()

        self.length = ''
        self.width = ''
        self.loss_min = ''
        self.loss_max = ''
        self.orient_start = ''
        self.orient_end = ''
        self.number = ''
        self.kbd = ''
        self.safe_pressure = ''
        self.time_limit = ''
        self.safe_pressure_persent = ''
        self.method_id = ''

    @classmethod
    def from_xml(cls, xml_item):
        """
        create defect item from existing xml element
        """
        obj = cls()
        obj.fill_from_xml(xml_item)

        obj.length = to_int(xml_item.attrib[Item.field_length])  # mm
        obj.width = to_int(xml_item.attrib[Item.field_width])  # mm
        obj.loss_min = to_int(xml_item.attrib[Item.field_loss_min])  # % or mm
        obj.loss_max = to_int(xml_item.attrib[Item.field_loss_max])  # % or mm
        obj.orient_start = xml_item.attrib[Item.field_orient_start]  # hour
        obj.orient_end = xml_item.attrib[Item.field_orient_end]  # hour
        obj.number = xml_item.attrib[Item.field_number]
        obj.kbd = xml_item.attrib[Item.field_kbd]
        obj.safe_pressure = xml_item.attrib[Item.field_safe_pressure]
        obj.time_limit = xml_item.attrib[Item.field_time_limit]
        obj.safe_pressure_persent = xml_item.attrib[Item.field_safe_pressure_persent]
        obj.method_id = xml_item.attrib[Item.field_method_id]

        return obj

    def as_csv_row(self, infotech, with_navigation=False):
        """
        return list of defect field values for csv string
        """
        method = ''
        if self.method_id in infotech.obj_dict:
            method = infotech.obj_dict[self.method_id]

        columns = [
          "{}".format(self.length),
          "{}".format(self.width),
          "{}".format(self.loss_min),
          "{}".format(self.loss_max),
          self.orient_start.replace('.', ','),
          self.orient_end.replace('.', ','),
          self.kbd,
          self.safe_pressure,
          self.time_limit,
          self.safe_pressure_persent,
          method,
        ]
        base_columns = super(Item, self).as_csv_row(infotech, with_navigation=with_navigation)

        return [self.number] + base_columns[:2] + columns + base_columns[2:]

    def reverse(self, total_length, object_index):
        """
        reverse defect
        """
        self.dist = total_length - self.dist
        if self.length:
            self.dist -= int(self.length / 10)

        self.number = "{}".format(object_index)
        self.orient_start = reverse_orient(self.orient_start)
        self.orient_end = reverse_orient(self.orient_end)

    def add_xml_child(self, parent_node):
        """
        create and add xmml node of defect to parent xml node
        """
        node = ET.SubElement(parent_node, Item.xml_node_name)
        super(Item, self).base_xml(node)
        node.set(Item.field_length, "{}".format(self.length))
        node.set(Item.field_width, "{}".format(self.width))
        node.set(Item.field_loss_min, "{}".format(self.loss_min))
        node.set(Item.field_loss_max, "{}".format(self.loss_max))
        node.set(Item.field_orient_start, self.orient_start)
        node.set(Item.field_orient_end, self.orient_end)
        node.set(Item.field_number, self.number)
        node.set(Item.field_kbd, self.kbd)
        node.set(Item.field_safe_pressure, self.safe_pressure)
        node.set(Item.field_time_limit, self.time_limit)
        node.set(Item.field_safe_pressure_persent, self.safe_pressure_persent)
        node.set(Item.field_method_id, self.method_id)

        return node


class Section(InfotechSection):
    """
    <DEFECTS> xml section
    """
    item_attributes = DistItem.dist_attribs + [
      Item.field_length,
      Item.field_width,
      Item.field_loss_min,
      Item.field_loss_max,
      Item.field_orient_start,
      Item.field_orient_end,
      Item.field_number,
      DistItem.field_comment,
      Item.field_kbd,
      Item.field_safe_pressure,
      Item.field_time_limit,
      Item.field_safe_pressure_persent,
      Item.field_method_id,
    ] + DistItem.coords_attribs

    def __init__(self, infotech):
        super(Section, self).__init__(infotech, Item, 'DEFECTS')

    def as_csv(self, with_navigation=False):
        """
        dump defects section content as csv string
        """
        column_titles = [
          'Number',
          'Name', 'Distance',
          'Length', 'Width', 'Loss min', 'Loss max', 'Orient start', 'Orient end',
          'KBD', 'Safe pressure', 'Time limit', 'Press percent', 'Method',
          'Comment',
        ]
        return super(Section, self).as_csv_body('Defects table', column_titles, with_navigation=with_navigation)

    def danger_stats(self):
        """
        generate statistics about defects danger data
        """
        stats_values = {
          DistItem.field_typeobj: [],
          Item.field_method_id: [],
        }
        stats_counts = {
          Item.field_kbd: 0,
          Item.field_safe_pressure: 0,
          Item.field_time_limit: 0,
          Item.field_safe_pressure_persent: 0,
          Item.field_method_id: 0,
        }

        for item in self.items:
            is_kbd = False

            if item.kbd:
                stats_counts[Item.field_kbd] += 1
                is_kbd = True

            if item.safe_pressure:
                stats_counts[Item.field_safe_pressure] += 1
                is_kbd = True

            if item.time_limit:
                stats_counts[Item.field_time_limit] += 1
                is_kbd = True

            if item.safe_pressure_persent:
                stats_counts[Item.field_safe_pressure_persent] += 1
                is_kbd = True

            if item.method_id:
                stats_counts[Item.field_method_id] += 1
                is_kbd = True
                if item.method_id not in stats_values[Item.field_method_id]:
                    stats_values[Item.field_method_id].append(item.method_id)

            if is_kbd:
                if item.objtype not in stats_values[DistItem.field_typeobj]:
                    stats_values[DistItem.field_typeobj].append(item.objtype)

        return (stats_values, stats_counts)



"""
LINEOBJS section
"""
from .ordered_attrib import ET
from .base import Section as InfotechSection, DistItem
from .codes import Feature


REVERTED = {
  Feature.WATER_START: Feature.WATER_END,
  Feature.WATER_END: Feature.WATER_START,

  Feature.METALL_CASE_START: Feature.METALL_CASE_END,
  Feature.METALL_CASE_END: Feature.METALL_CASE_START,

  Feature.KOMPOS_CASE_START: Feature.KOMPOS_CASE_END,
  Feature.KOMPOS_CASE_END: Feature.KOMPOS_CASE_START,

  Feature.CASE_START: Feature.CASE_END,
  Feature.CASE_END: Feature.CASE_START,

  Feature.PRIGRUZ_START: Feature.PRIGRUZ_END,
  Feature.PRIGRUZ_END: Feature.PRIGRUZ_START,

  Feature.TURN_START: Feature.TURN_END,
  Feature.TURN_END: Feature.TURN_START,

  Feature.TURN_SEGM_START: Feature.TURN_SEGM_END,
  Feature.TURN_SEGM_END: Feature.TURN_SEGM_START,
}


class Item(DistItem):
    """
    object item from <LINEOBJS> xml section
    <PLOBJ
      IDTYPEOBJ="990006537229" ODOMETER="0" NAME_MARKER="M1" L_LCH="" REM=""
    />
    """
    xml_node_name = 'PLOBJ'
    field_name_marker = 'NAME_MARKER'
    field_piketag_km = 'L_LCH'

    def __init__(self):
        super(Item, self).__init__()
        self.marker_name = ''
        self.piketag_km = ''

    @classmethod
    def from_xml(cls, xml_item):
        """
        create lineobj item from existing xml element
        """
        obj = cls()
        obj.fill_from_xml(xml_item)

        obj.marker_name = xml_item.attrib[Item.field_name_marker]
        obj.piketag_km = xml_item.attrib[Item.field_piketag_km]

        return obj

    def as_csv_row(self, infotech, with_navigation=False):
        """
        return list of lineobj field values for csv string
        """
        columns = [self.marker_name, self.piketag_km]
        base_columns = super(Item, self).as_csv_row(infotech, with_navigation=with_navigation)

        return base_columns[:2] + columns + base_columns[2:]

    def add_xml_child(self, parent_node):
        """
        create and add lineobj xml node of object to parent xml node
        """
        node = ET.SubElement(parent_node, Item.xml_node_name)
        super(Item, self).base_xml(node)
        node.set(Item.field_name_marker, self.marker_name)
        node.set(Item.field_piketag_km, self.piketag_km)

        return node

    def reverse(self, total_length, object_index):
        """
        reverse line object
        """
        super(Item, self).reverse(total_length, object_index)
        if self.objtype in REVERTED:
            self.objtype = REVERTED[self.objtype]


class Section(InfotechSection):
    """
    <LINEOBJS> xml section
    """
    section_tag = 'LINEOBJS'

    item_attributes = DistItem.dist_attribs + [
      Item.field_name_marker,
      Item.field_piketag_km,
      DistItem.field_comment,
    ] + DistItem.coords_attribs

    def __init__(self, infotech):
        super(Section, self).__init__(infotech, Item, Section.section_tag)

    def as_csv(self, with_navigation=False):
        """
        dump lineobj section content as csv string
        """
        column_titles = [
          'Name', 'Distance', 'Marker', 'Piketag', 'Comment'
        ]
        return super(Section, self).as_csv_body('Line objects table', column_titles, with_navigation=with_navigation)



# coding: utf-8
"""
Infotech staff
"""
from StringIO import StringIO, StringIO as BytesIO  # pylint: disable=reimported
from .ordered_attrib import ET


def reverse_orient(orient):
    """
    reverse orientation
    """
    if not orient:
        return orient

    hours = 12.0 - float(orient.replace(',', '.'))
    if hours == 12.0:
        hours = 0.0

    return "{0:0.1f}".format(hours).replace('.', ',')


def numerate(items, from_index):
    """
    enumerate items from given from_index
    """
    for item in items:
        item.number = "{}".format(from_index)
        from_index += 1


def indent(elem, level=0, ident_item="    "):
    """
    http://effbot.org/zone/element-lib.htm#prettyprint
    in-place prettyprint formatter
    """
    i = "\n" + level*ident_item
    if len(elem):  # pylint: disable=len-as-condition
        if not elem.text or not elem.text.strip():
            elem.text = i + ident_item
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:  # pylint: disable=redefined-argument-from-local
            indent(elem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


class Infotech(object):
    """
    Infotech xml export
    """
    typobj_section = 'TYPEOBJS'
    typobj_item = 'TYPEOBJ'
    typobj_id = 'IDTYPEOBJ'
    typobj_title = 'TITLE'

    gps_sys_code = '0'
    gps_sys_name = 'WGS-84'

    template = """<?xml version="1.0" encoding="windows-1251"?>
<IPL_INSPECT>
<TYPEOBJS></TYPEOBJS>
<DEFECTS></DEFECTS>
<LINEOBJS></LINEOBJS>
<WELDS></WELDS>
<PIGPASS></PIGPASS>
</IPL_INSPECT>
"""

    def __init__(self, codepage="windows-1251"):
        self.xml = ET.parse(StringIO(self.template))
        self.obj_dict = {}
        self.codepage = codepage
        self.is_navigate = None

        from . import defect, weld, lineobj, pigpass
        self.welds = weld.Section(self)
        self.defects = defect.Section(self)
        self.lineobjects = lineobj.Section(self)
        self.pigpass = pigpass.Section(self)

    def __unicode__(self):
        output = BytesIO()
        self.xml.write(output, encoding=self.codepage)

        return output.getvalue()

    def __str__(self):
        return self.__unicode__()

    @classmethod
    def from_file(cls, file_name):
        """
        load data from xml file
        """
        obj = cls()
        obj.xml = ET.parse(file_name)
        sect = obj.xml.getroot().find(obj.typobj_section)
        obj.obj_dict = {typ.attrib[obj.typobj_id]: typ.find(obj.typobj_title).text for typ in sect}

        return obj

    def rebuild_typeobjs(self):
        """
        generate new content for TYPEOBJS section of xml, based on obj_dict data
        """
        section = self.xml.getroot().find(self.typobj_section)
        section.clear()

        if self.is_navigate:
            node = ET.SubElement(section, self.typobj_item)
            node.set(self.typobj_id, self.gps_sys_code)
            title = ET.SubElement(node, self.typobj_title)
            title.text = self.gps_sys_name

        for key, val in self.obj_dict.items():
            node = ET.SubElement(section, self.typobj_item)
            node.set(self.typobj_id, key)
            title = ET.SubElement(node, self.typobj_title)
            title.text = val

    def total_dist(self):
        """
        return length of inspection
        """
        return self.welds.items[-1].end()

    def start_dist(self):
        """
        return start distance of inspection
        """
        return self.welds.items[0].dist

    def reverse(self):
        """
        reverse vector of objects and return string dump of updated xml
        """
        total_length = self.total_dist()
        self.lineobjects.reverse(total_length)
        self.welds.reverse(total_length)
        self.defects.reverse(total_length)

        return "{}".format(self)

    def fix(self):
        """
        repair umdp-1400 data in PIGPASS section
        """
        umdp = u'УМДП-1400'  # encode(self.codepage)
        umdp_key = None

        for key, val in self.obj_dict.items():
            if val == umdp:
                umdp_key = key
                break

        if not umdp_key:
            return "{}".format(self)

        for pig in self.xml.getroot().find('PIGPASS'):
            if pig.attrib.get('IDTYPEOBJ', None) == umdp_key:
                pig.set('MANUFACT_DATE', '2017')
                pig.set('PIGTYPE', '990004033563')
                pig.set('OBSLTYPE', '2')

        return "{}".format(self)

    def join(self, file_list):
        """
        join several xml files and connect tubes
        """
        from .codes import Tube, NAME
        is_updated = False

        for item in file_list:
            current_length = self.total_dist()
            try:
                self.welds.add_tube(Tube.UNKNOWN, current_length, int(item))
                if Tube.UNKNOWN not in self.obj_dict:
                    self.obj_dict[Tube.UNKNOWN] = NAME[Tube.UNKNOWN]

                is_updated = True

            except ValueError:

                try:
                    info = Infotech.from_file(item)
                except IOError as exc:
                    return str(exc)

                for key in info.obj_dict:
                    if len(key) > 2:
                        if key not in self.obj_dict:
                            self.obj_dict[key] = info.obj_dict[key]

                start = info.start_dist()

                self.welds.add_from_xml(info.welds, current_length, start)
                self.defects.add_from_xml(info.defects, current_length, start)
                self.lineobjects.add_from_xml(info.lineobjects, current_length, start)

                is_updated = True

        if is_updated:
            numerate(self.welds.items, 1)
            numerate(self.defects.items, 1)

            self.rebuild_typeobjs()
            self.welds.rebuild_xml()
            self.defects.rebuild_xml()
            self.lineobjects.rebuild_xml()

        indent(self.xml.getroot(), level=0)

        return "{}".format(self)



"""
Base classes for Infotech sections and items
"""
import csv
from StringIO import StringIO
from . import indent


def to_int(text):
    """
    text to int
    """
    try:
        return int(text)
    except ValueError:
        return ''


class AbstractItem(object):  # pylint: disable=too-few-public-methods
    """
    abstract class for xml item
    """
    def is_valid(self):
        """
        check for valid field values
        """
        return bool(self)


class DistItem(AbstractItem):
    """
    abstract class for item on dist
    """
    field_typeobj = 'IDTYPEOBJ'
    field_odometer = 'ODOMETER'
    field_comment = 'REM'
    # not required fields
    field_coord_longitude = 'L'
    field_coord_latitude = 'B'
    field_coord_height = 'H'
    field_coord_system = 'SRID'

    dist_attribs = [
      field_typeobj,
      field_odometer,
    ]

    coords_attribs = [
      field_coord_longitude,
      field_coord_latitude,
      field_coord_height,
      field_coord_system,
    ]

    def __init__(self):
        self.objtype = None
        self.dist = None
        self.comment = ''

        self.coord_longitude = None
        self.coord_latitude = None
        self.coord_height = None
        self.coord_system = None

    def fill_from_xml(self, xml_item):
        """
        fill item data from existing xml element
        """
        self.objtype = xml_item.attrib[DistItem.field_typeobj]
        self.dist = to_int(xml_item.attrib[DistItem.field_odometer])
        self.comment = xml_item.attrib[DistItem.field_comment]

        self.coord_longitude = xml_item.attrib.get(DistItem.field_coord_longitude, None)
        self.coord_latitude = xml_item.attrib.get(DistItem.field_coord_latitude, None)
        self.coord_height = xml_item.attrib.get(DistItem.field_coord_height, None)
        self.coord_system = xml_item.attrib.get(DistItem.field_coord_system, None)

    def as_csv_row(self, infotech, with_navigation=False):
        """
        return list of item fields for csv output
        """
        suffix = []
        if with_navigation:
            suffix = [
              self.coord_longitude,
              self.coord_latitude,
              self.coord_height,
              self.coord_system,
            ]

        return [
          infotech.obj_dict[self.objtype],
          "{}".format(self.dist),
          self.comment,
        ] + suffix

    def reverse(self, total_length, _object_index):
        """
        reverse object distance
        """
        self.dist = total_length - self.dist

    def base_xml(self, node):
        """
        set base attributes to xml node of object
        """
        node.set(DistItem.field_typeobj, self.objtype)
        node.set(DistItem.field_odometer, "{}".format(self.dist))
        node.set(DistItem.field_comment, self.comment)

        if self.coord_system is not None:
            node.set(DistItem.field_coord_longitude, self.coord_longitude)
            node.set(DistItem.field_coord_latitude, self.coord_latitude)
            node.set(DistItem.field_coord_height, self.coord_height)
            node.set(DistItem.field_coord_system, self.coord_system)


class Section(object):
    """
    abstract class for xml section
    """
    item_attributes = None

    def __init__(self, infotech, cls, section):
        self.infotech = infotech
        self.cls = cls
        self.section = section
        self._items = None

    @property
    def items(self):
        """
        lazy restore list of section's items
        """
        if self._items is None:
            self._items = []
            for obj in self.infotech.xml.getroot().find(self.section):
                item = self.cls.from_xml(obj)
                if item.is_valid():
                    self._items.append(item)

        return self._items

    def as_csv_body(self, title, column_titles, with_navigation=False):
        """
        dump section as csv string
        """
        output = StringIO()
        writer = csv.writer(output, delimiter=';', lineterminator='\n')

        writer.writerow([title] + ['' for _tmp in column_titles][1:])
        writer.writerow(column_titles)
        for item in self.items:
            writer.writerow([
              field.encode(self.infotech.codepage)
              for field in item.as_csv_row(self.infotech, with_navigation=with_navigation)
            ])

        content = output.getvalue()
        output.close()

        return content

    def rebuild_xml(self):
        """
        generate new xml data based on current section data
        """
        section = self.infotech.xml.getroot().find(self.section)
        section.clear()

        for item in self.items:
            node = item.add_xml_child(section)
            if self.item_attributes:
                node.ordered_attributes = self.item_attributes

    def reverse(self, total_length):
        """
        reverse vector of objects and modify self.infotech.xml
        """
        self._items = reversed(self.items)
        section = self.infotech.xml.getroot().find(self.section)
        section.clear()
        object_index = 1

        for item in self.items:
            item.reverse(total_length, object_index)
            item.add_xml_child(section)
            object_index += 1

        indent(section, level=1)

    def add_from_xml(self, xml_section, at_dist, dist_shift):
        """
        add objects from another infotech section from given dist
        """
        items = self.items
        for item in xml_section.items:
            item.dist = item.dist - dist_shift + at_dist
            items.append(item)



import setuptools


long_description = """
# oeg.infotech.xml

Library for OrgEnergoGaz Infotech xml processing
"""

setuptools.setup(
    name = 'oeg_infotech',
    version = '1.1',
    author = 'Vitaly Bogomolov',
    author_email = 'mail@vitaly-bogomolov.ru',
    description = 'Library for OrgEnergoGaz Infotech xml processing',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url = 'https://github.com/vb64/oeg.infotech.xml',
    packages = ['oeg_infotech'],
    download_url = 'https://github.com/vb64/oeg.infotech.xml/archive/v1.1.tar.gz',
    keywords = ['python', 'OrgEnergoGaz', 'Infotech', 'xml'],
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""
https://stackoverflow.com/questions/2741480/can-elementtree-be-told-to-preserve-the-order-of-attributes
"""
import xml.etree.ElementTree as ET


def get_items(elem):
    """
    sort attribs by given order, if set
    """
    if not hasattr(elem, 'ordered_attributes'):
        return sorted(elem.items())

    items = dict(elem.items())
    return [(attr, items[attr]) for attr in elem.ordered_attributes if attr in items]


def _serialize_xml(write, elem, encoding, qnames, namespaces, short_empty_elements=None):
    """
    monkey patch for dump ordered attributes for xml element
    """
    tag = elem.tag
    text = elem.text
    if tag is ET.Comment:
        write("<!--%s-->" % ET._encode(text, encoding))
    elif tag is ET.ProcessingInstruction:
        write("<?%s?>" % ET._encode(text, encoding))
    else:
        tag = qnames[tag]
        if tag is None:
            if text:
                write(ET._escape_cdata(text, encoding))
            for e in elem:
                _serialize_xml(write, e, encoding, qnames, None)
        else:
            write("<" + tag)
            items = get_items(elem)
            if items or namespaces:
                if namespaces:
                    for v, k in sorted(namespaces.items(),
                                       key=lambda x: x[1]):  # sort on prefix
                        if k:
                            k = ":" + k
                        write(" xmlns%s=\"%s\"" % (
                            k.encode(encoding),
                            ET._escape_attrib(v, encoding)
                            ))
                for k, v in items:
                    if isinstance(k, ET.QName):
                        k = k.text
                    if isinstance(v, ET.QName):
                        v = qnames[v.text]
                    else:
                        v = ET._escape_attrib(v, encoding)
                    write(" %s=\"%s\"" % (qnames[k], v))
            if text or len(elem):
                write(">")
                if text:
                    write(ET._escape_cdata(text, encoding))
                for e in elem:
                    _serialize_xml(write, e, encoding, qnames, None)
                write("</" + tag + ">")
            else:
                write(" />")
    if elem.tail:
        write(ET._escape_cdata(elem.tail, encoding))


ET._serialize_xml = _serialize_xml
ET._serialize["xml"] = _serialize_xml  # dump root elem
